/*------------------------------------------------------------------------------
 *
 *       Filename:  ar.c
 *
 *    Description:  Activity recognition on streaming sensor data.
 *
 *        Version:  2.0
 *        Created:  05/07/2012 06:46:51 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *        Authors:  Diane J. Cook, cook@eecs.wsu.edu
 *   Organization:  Washington State University
 * ---------------------------------------------------------------------------*/

#include "ar.h"


/* The command line syntax is ar [configfile].  The data file and program
   parameters are specified in the config file. If no config filename is
   provided then the default file of ar.config is assumed. */
int main(int argc, char *argv[])
{
   char fname[MAXSTR];
   int i;

   printf("AR - Reading the config file...");
   if (argc > 1)
      ReadConfig(argv[1]);
   else
   {
      strcpy(fname, "ar.config");
      ReadConfig(fname);
   }
   printf("done\n");

   if (argc > 2)                    // Filename specified by user
   {
      printf("Reading user specified data file names...");
      for (i=2; i<argc; i++)        // Process all files listed on command line
      {
         if (NumFiles == 0)
            filenames = (char **) malloc(sizeof(char *));
         else filenames =
            (char **) realloc(filenames, (NumFiles + 1) * sizeof(char *));
	 if (filenames == NULL)
	 {
	    fprintf(stderr, "Memory error in main\n");
	    exit(1);
	 }
         filenames[NumFiles] = (char *) malloc(MAXSTR * sizeof(char));
	 if (filenames[NumFiles] == NULL)
	 {
	    fprintf(stderr, "Memory error in main\n");
	    exit(1);
	 }
         strcpy(filenames[NumFiles], argv[i]);
         NumFiles++;
      }
      printf("done\n\n");
   }

   if (OutputLevel > 0)
       PrintParms();

   if (Mode == TRAIN) // TRAIN mode
   {
      printf("Entering training mode\n");

      // Compute the window probabilities
      if (OutputLevel > 0)
         printf("   Computing probabilities to determine window size...");

      // Compute activity probability distribution for alternative window sizes
      ComputeProbActWind();

      // Compute activity probability distribution for each sensor
      ComputeProbSenAct();

      if (OutputLevel > 0)
         printf("done\n");

      // Compute mutual information between the sensors
      if (Weight > 1)
      {
         if (OutputLevel > 0)
            printf("Computing mutual information between the sensors...");
         ComputeMI();
         if (OutputLevel > 0)
            printf("done\n");
      }

      ReadData();             // Read the data
      // create directory to store the model files and the final config file
      mkdir(modelfname, S_IRWXU | S_IRWXG | S_IRWXO);
      TrainModel();           // Learn and save the model
      SaveParams();           // Save the program parameters to a file
   }
   else // TEST mode
   {
      if ((factwind == NULL) || (fsenact == NULL))
      {
         fprintf(stderr,
	    "Factwind and fsenact values need to be provided for testing\n");
         exit(1);
      }
      printf("Reading data...");
      ReadData();             // Read the data
      printf("done\n");
      TestModel();            // Load the activity model and test the data
      printf("accuracy is %f\n", (float) numright / (float) NumData);
   }
   FreeMemory();

   return(1);
}


/* Read each line of the config file and set program parameters accordingly. */
void ReadConfig(char *name)
{
   FILE *fp;
   char *cptr=NULL, *buffer=NULL, arg[MAXSTR];
   int i, j, pos, length;

   fp = fopen(name, "r");
   if (fp == NULL)
   {
      printf("Unable to open the configuration file: %s\n", name);
      exit(1);
   }

   buffer = (char *) malloc(BUFFERSIZE * sizeof(char));
   if (buffer == NULL)
   {
      fprintf(stderr, "Memory error in ReadConfig\n");
      exit(1);
   }

   cptr =  fgets(buffer, BUFFERSIZE, fp);  // Read next line
   while (cptr != NULL)                // Process each line of the config file
   {
      buffer = TrimStr(buffer);
      length = (int) strlen(buffer);
      sscanf(buffer, "%s", arg);
      if (strcmp(arg, "output") == 0)  // Specify output reporting level
         sscanf(buffer, "%s %d\n", arg, &OutputLevel);
      else if (strcmp(arg, "data") == 0)  // Data file name(s)
      {
	 pos = strlen(arg) + 1;
	 if (pos >= (int) strlen(buffer))
         {
            printf("No data file specified\n");
            exit(1);
         }
	 pos = ReadString(buffer, pos, arg, length);
	 while (pos >= 0)
         {
            if (NumFiles == 0)
	       filenames = (char **) malloc(sizeof(char *));
            else filenames =
               (char **) realloc(filenames, (NumFiles + 1) * sizeof(char *));
	    if (filenames == NULL)
	    {
	       fprintf(stderr, "Memory error in ReadConfig\n");
	       exit(1);
	    }
            filenames[NumFiles] = (char *) malloc(MAXSTR * sizeof(char));
	    if (filenames[NumFiles] == NULL)
	    {
	       fprintf(stderr, "Memory error in ReadConfig\n");
	       exit(1);
	    }
            strcpy(filenames[NumFiles], arg);
	    NumFiles++;
	    pos = ReadString(buffer, pos, arg, length);
         }
      }
      else if (strcmp(arg, "mode") == 0)   // Train or test mode
         sscanf(buffer, "%s %d\n", arg, &Mode);
      else if (strcmp(arg, "weight") == 0)  // How to weight events in window
         sscanf(buffer, "%s %d\n", arg, &Weight);
      else if (strcmp(arg, "ignoreother") == 0) // Ignore OtherActivity
         IgnoreOther = TRUE;
      else if (strcmp(arg, "model") == 0)   // Specify model directory name
      {
	 modelfname = (char *) malloc(MAXSTR * sizeof(char));
         sscanf(buffer, "%s %s\n", arg, modelfname);
      }
      else if (strcmp(arg, "sensor") == 0)  // List of sensor names
      {
	 pos = strlen(arg) + 1;
	 pos = ReadString(buffer, pos, arg, length);
	 while (pos >= 0)
         {
            if (NumSensors == 0)
	    {
               sensornames = (char **) malloc(sizeof(char *));
	       sensortimes = (int *) malloc(sizeof(int));
	    }
            else
	    {
	       sensornames = (char **)
	          realloc(sensornames, (NumSensors+1) * sizeof(char *));
	       sensortimes = (int *)
	          realloc(sensortimes, (NumSensors+1) * sizeof(int));
	    }
	    if ((sensornames == NULL) || (sensortimes == NULL))
	    {
	       fprintf(stderr, "Memory error in ReadConfig\n");
	       exit(1);
	    }
            sensornames[NumSensors] = (char *) malloc(MAXSTR * sizeof(char));
	    if (sensornames[NumSensors] == NULL)
	    {
	       fprintf(stderr, "Memory error in ReadConfig\n");
	       exit(1);
	    }
            strcpy(sensornames[NumSensors], arg);
	    sensortimes[NumSensors] = 86400;
            NumSensors++;
	    pos = ReadString(buffer, pos, arg, length);
         }
      }
      else if (strcmp(arg, "activity") == 0)  // List of activity labels
      {
	 pos = strlen(arg) + 1;
	 pos = ReadString(buffer, pos, arg, length);
	 while (pos >= 0)
         {
            if (NumActivities == 0)
	    {
               activitynames = (char **) malloc(sizeof(char *));
	       classfreq = (int *) malloc(sizeof(int));
	    }
            else
	    {
	       activitynames = (char **) realloc(activitynames,
                  (NumActivities+1) * sizeof(char *));
	       classfreq = (int *) realloc(classfreq,
	          (NumActivities+1) * sizeof(int));
	    }
	    if ((activitynames == NULL) || (classfreq == NULL))
	    {
	       fprintf(stderr, "Memory error in ReadConfig\n");
	       exit(1);
	    }
            activitynames[NumActivities] =
               (char *) malloc(MAXSTR * sizeof(char));
	    if (activitynames[NumActivities] == NULL)
	    {
	       fprintf(stderr, "Memory error in ReadConfig\n");
	       exit(1);
	    }
            strcpy(activitynames[NumActivities], arg);
	    classfreq[NumActivities] = 0;
            NumActivities++;
	    pos = ReadString(buffer, pos, arg, length);
         }
      }
      else if (strcmp(arg, "fsenact") == 0)  // Activity likelihood for sensors
      {
         i = 0;
	 pos = strlen(arg) + 1;
	 pos = ReadString(buffer, pos, arg, length);
	 while (pos >= 0)
         {
            if (i == 0)
               fsenact = (int *) malloc(sizeof(int));
            else fsenact = (int *) realloc(fsenact, (i+1) * sizeof(int));
	    if (fsenact == NULL)
	    {
	       fprintf(stderr, "Memory error in ReadConfig\n");
	       exit(1);
	    }
            fsenact[i] = atoi(arg);
            i++;
	    pos = ReadString(buffer, pos, arg, length);
         }
      }
      else if (strcmp(arg, "factwind") == 0) // Activity probs for window sizes
      {
         i = 0;
	 pos = strlen(arg) + 1;
	 pos = ReadString(buffer, pos, arg, length);
	 while (pos >= 0)
         {
            if (i == 0)
               factwind = (int *) malloc(sizeof(int));
            else factwind = (int *) realloc(factwind, (i+1) * sizeof(int));
	    if (factwind == NULL)
	    {
	       fprintf(stderr, "Memory error in ReadConfig\n");
	       exit(1);
	    }
            factwind[i] = atoi(arg);
            i++;
	    pos = ReadString(buffer, pos, arg, length);
         }
      }
      else if (strcmp(arg, "mi") == 0)  // Sensor mutual information
      {
	 pos = strlen(arg) + 1;
	 pos = ReadString(buffer, pos, arg, length);
         MI = (float **) malloc(NumSensors * sizeof(float *));
	 if (MI == NULL)
	 {
	    fprintf(stderr, "Memory error in ReadConfig\n");
	    exit(1);
	 }
         for (i=0; i<NumSensors; i++)
         {
            MI[i] = (float *) malloc(NumSensors * sizeof(float));
	    if (MI[i] == NULL)
	    {
	       fprintf(stderr, "Memory error in ReadConfig\n");
	       exit(1);
	    }
            for (j=0; j<NumSensors; j++)
            {
               MI[i][j] = (float) atof(arg);
	       pos = ReadString(buffer, pos, arg, length);
	    }
         }
      }
      else fprintf(stderr, "Do not recognize config option %s\n", arg);
      cptr =  fgets(buffer, BUFFERSIZE, fp);  // Read next line
   }

   fclose(fp);
   free(buffer);
}


/* Remove white space from the end of a line. */
char *TrimStr(char *str)
{
   int i, length;

   length = (int) strlen(str);
   i = length - 1;
   while ((i > 0) && (isspace(str[i]) != 0))
      i--;
   str[i+1] = '\0';
   length = (int) strlen(str);

   return(str);
}


/* Read the next string from a line buffer and update the current
   position in the buffer. */
int ReadString(char *str, int pos, char *arg, int length)
{
   int i;

   i = pos;
                    // Add to string until reach end of line or white space
   while ((i < length) && (isspace(str[i]) == 0))
   {
      arg[i-pos] = str[i];
      i++;
   }
   arg[i-pos] = '\0';
   if (i > length)
      return(-1);
   else return(i+1);
}


/* Print program parameter values */
void PrintParms()
{
   int i;

   printf("*****************************\n");
   printf("*** AR Program Parameters ***\n\n");
   if (Mode == 0)
      printf("Mode: Training mode\n");
   else printf("Mode: Testing mode\n");

   if (Weight == 0)
      printf("Weight: No sensor event weighting\n");
   else if (Weight == 1)
      printf("Weight: Time sensor event weighting\n");
   else if (Weight == 2)
      printf("Weight: Mutual information sensor event weighting\n");
   else printf("Weight: Time and mutual information sensor event weighting\n");

   if (modelfname != NULL)
      printf("Model file: %s\n", modelfname);

   printf("Sensor names [%d]:", NumSensors);
   for (i=0; i<NumSensors; i++)
      printf(" %s", sensornames[i]);
   printf("\n");

   if (Mode != TRAIN)
   {
      printf("Activity labels: [%d]", NumActivities);
      for (i=0; i<NumActivities; i++)
         printf(" %s", activitynames[i]);
      printf("\n");
   }
   printf("*****************************\n\n");
}


/* Save the parameters for use during testing */
void SaveParams()
{
   FILE *fp;
   char str[MAXSTR];
   int i, j;

   sprintf(str, "%s/%s", modelfname, "model.config");
   fp = fopen(str, "w");
   if (fp == NULL)
   {
      printf("Unable to open %s for writing\n", str);
      exit(-1);
   }

   // sensor names
   fprintf(fp, "sensor ");
   for (i=0; i<NumSensors; i++)
      fprintf(fp, "%s ", sensornames[i]);
   fprintf(fp, "\n");

   // activity names
   fprintf(fp, "activity ");
   for (i=0; i<NumActivities; i++)
      fprintf(fp, "%s ", activitynames[i]);
   fprintf(fp, "\n");

   // fsenact
   fprintf(fp, "fsenact ");
   for (i=0; i<NumSensors; i++)
      fprintf(fp, "%d ", fsenact[i]);
   fprintf(fp, "\n");

   // factwind
   fprintf(fp, "factwind ");
   for (i=0; i<NumActivities; i++)
      fprintf(fp, "%d ", factwind[i]);
   fprintf(fp, "\n");

   // weight
   fprintf(fp, "weight %d\n", Weight);

   // mi
   if (Weight > 1)
   {
      fprintf(fp, "mi ");
      for (i=0; i<NumSensors; i++)
         for (j=0; j<NumSensors; j++)
            fprintf(fp, "%.6f ", MI[i][j]);
      fprintf(fp, "\n");
   }

   // model directory
   fprintf(fp, "model %s\n", modelfname);

   // test mode
   fprintf(fp, "mode 1\n");

   // ignore other activity
   if (IgnoreOther == TRUE)
      fprintf(fp, "ignoreother\n");

   fclose(fp);

   if (OutputLevel > 0)
      printf("Parameters saved to file.\n");
}


/* Find the index of a specified activity from the activity list */
int FindActivity(char *name)
{
   int i;

   for (i=0; i<NumActivities; i++)
      if (strcmp(name, activitynames[i]) == 0)
         return i;

   if (NumActivities == 0)
   {
      activitynames = (char **) malloc(sizeof(char *));
      if (activitynames == NULL)
      {
         fprintf(stderr, "Memory error in FindActivity\n");
         exit(1);
      }
      actwind = (int **) malloc(sizeof(int *));
      if (actwind == NULL)
      {
         fprintf(stderr, "Memory error in FindActivity\n");
         exit(1);
      }
      classfreq = (int *) malloc(sizeof(int));
      if (classfreq == NULL)
      {
         fprintf(stderr, "Memory error in FindActivity\n");
         exit(1);
      }
   }
   else
   {
      activitynames =
          (char **) realloc(activitynames, (NumActivities+1) * sizeof(char *));
      if (activitynames == NULL)
      {
         fprintf(stderr, "Memory error in FindActivity\n");
         exit(1);
      }
      actwind = (int **) realloc(actwind, (NumActivities+1) * sizeof(int *));
      if (actwind == NULL)
      {
         fprintf(stderr, "Memory error in FindActivity\n");
         exit(1);
      }
      classfreq = (int *) realloc(classfreq, (NumActivities+1) * sizeof(int));
      if (classfreq == NULL)
      {
         fprintf(stderr, "Memory error in FindActivity\n");
         exit(1);
      }
   }	
   activitynames[NumActivities] = (char *) malloc(MAXSTR * sizeof(char));
   if (activitynames[NumActivities] == NULL)
   {
      fprintf(stderr, "Memory error in FindActivity\n");
      exit(1);
   }
   strcpy(activitynames[NumActivities], name);

   actwind[NumActivities] = (int *) calloc(NUMWINDOWS, sizeof(int));
   if (actwind[NumActivities] == NULL)
   {
      fprintf(stderr, "Memory error in FindActivity\n");
      exit(1);
   }

   classfreq[NumActivities] = 0;
   NumActivities++;

   return(NumActivities-1);
}


/* Find the index of a specified sensor from the sensor list */
int FindSensor(char *name)
{
   int i;

   for (i=0; i<NumSensors; i++)
      if (strcmp(name, sensornames[i]) == 0)
         return(i);
   printf("Could not find the sensor %s\n", name);
   exit(-1);
}


/* Compute the probability distribution over activity labels for alternative
   window sizes.  Determine the most likely window size given an activity.   */
void ComputeProbActWind()
{
   FILE *fp;
   char *cptr=NULL;
   char buffer[BUFFERSIZE], sdate[MAXSTR], stime[MAXSTR], sensorid[MAXSTR];
   char newsensorid[MAXSTR], sensorstatus[MAXSTR], alabel[MAXSTR];
   char prevlabel[MAXSTR];
   int cnt, i, j, anum, max, numfile;

   for (numfile=0; numfile<NumFiles; numfile++)  // Process each data file
   {
      fp = fopen(filenames[numfile], "r");
      if (fp == NULL)
      {
         printf("Unable to open data file %s\n", filenames[numfile]);
         exit(1);
      }

      cptr = fgets(buffer, BUFFERSIZE, fp);
      sscanf(buffer, "%s %s %s %s %s %s",
             sdate, stime, sensorid, newsensorid, sensorstatus, alabel);

      strcpy(prevlabel, alabel);            // Keep track of previous activity
      cnt = 0;
      while (cptr != NULL)  // Process each sensor event in the file
      {
         sscanf(buffer, "%s %s %s %s %s %s",
                sdate, stime, sensorid, newsensorid, sensorstatus, alabel);

	 if ((IgnoreOther == FALSE) || (strcmp(alabel, "Other_Activity") != 0))
	 {
            // Store activity probability distribution for
            // four alternative window sizes: 5, 10, 20, and 30.
	    // Determine ideal window size by looking at actual length of
	    // activities in sample data.
            if (strcmp(prevlabel, alabel) != 0)
            {
               anum = FindActivity(prevlabel);
               if (cnt <= 5)
                  actwind[anum][0]++;
               else if (cnt > 5 && cnt <= 10)
                  actwind[anum][1]++;
               else if (cnt > 10 && cnt <= 20)
                  actwind[anum][2]++;
               else if (cnt > 20)
                  actwind[anum][3]++;
               cnt = 0;
               strcpy(prevlabel, alabel);
            }
            else cnt++;
         }

         cptr = fgets(buffer, BUFFERSIZE, fp);
      }
      fclose(fp);
   }

   // Identify the window size with the greatest probability for each activity
   factwind = (int *) calloc(NumActivities, sizeof(int));
   if (factwind == NULL)
   {
      fprintf(stderr, "Memory error in ComputeProbActWind\n");
      exit(1);
   }
   for (i=0; i<NumActivities; i++)
   {
      max = 0;
      for (j=0; j<NUMWINDOWS; j++)
      {
         if (actwind[i][j] > max)
         {
            factwind[i] = j;
            max = actwind[i][j];
         }
      }
   }
}


/* Compute the probability distribution of the activities given the sensor and
   identify the most likely activity given the sensor */
void ComputeProbSenAct()
{
   FILE *fp;
   char *cptr=NULL;
   char buffer[BUFFERSIZE], sdate[MAXSTR], stime[MAXSTR], sensorid[MAXSTR];
   char newsensorid[MAXSTR], sensorstatus[MAXSTR], alabel[MAXSTR];
   int i, j, snum, anum, max, numfile;

   senact = (int **) malloc(NumSensors * sizeof(int *));
   if (senact == NULL)
   {
      fprintf(stderr, "Memory error in ComputeProbActWind\n");
      exit(1);
   }
   for (i=0; i<NumSensors; i++)
   {
      senact[i] = (int *) calloc(NumActivities, sizeof(int));
      if (senact[i] == NULL)
      {
         fprintf(stderr, "Memory error in ComputeProbSenAct\n");
         exit(1);
      }
   }

   for (numfile=0; numfile<NumFiles; numfile++)  // Process each data file
   {
      fp = fopen(filenames[numfile], "r");
      if (fp == NULL)
      {
         printf("Unable to open %s\n", filenames[numfile]);
         exit(1);
      }

      cptr = fgets(buffer, BUFFERSIZE, fp);
      while (cptr != NULL)  // Process each sensor event in the file
      {
         sscanf(buffer, "%s %s %s %s %s %s",
                   sdate, stime, sensorid, newsensorid, sensorstatus, alabel);
         if (strcmp(alabel, "none") != 0)
         {
            snum = FindSensor(newsensorid);
            anum = FindActivity(alabel);
            senact[snum][anum]++;
         }
         cptr = fgets(buffer, BUFFERSIZE, fp);
      }
      fclose(fp);
   }

   // Identify most likely activity associated with each sensor
   fsenact = (int *) calloc(NumSensors, sizeof(int));
   if (fsenact == NULL)
   {
      fprintf(stderr, "Memory error in ComputeProbActWind\n");
      exit(1);
   }
   for (i=0; i<NumSensors; i++)
   {
      max = 0;
      for (j=0; j<NumActivities; j++)
      {
         if (senact[i][j] > max)
         {
            fsenact[i] = j;
            max = senact[i][j];
         }
      }
   }
}


/* Compute the mutual information between sensors by determining the
   probability of co-occurence of every pair of sensors */
void ComputeMI()
{
   FILE *fp;
   char *cptr=NULL;
   char buffer[BUFFERSIZE], sdate[MAXSTR], stime[MAXSTR], sensorid[MAXSTR];
   char newsensorid[MAXSTR], sensorstatus[MAXSTR], alabel[MAXSTR];
   int i, j, numfile, presen, cursen;
   float total;

   MI = (float **) malloc(NumSensors * sizeof(float *));
   if (MI == NULL)
   {
      fprintf(stderr, "Memory error in ComputeMI\n");
      exit(1);
   }
   for (i=0; i<NumSensors; i++)
   {
      MI[i] = (float *) calloc(NumSensors, sizeof(float));
      if (MI[i] == NULL)
      {
         fprintf(stderr, "Memory error in ComputeMI\n");
         exit(1);
      }
   }

   for (numfile=0; numfile<NumFiles; numfile++)  // Process each data file
   {
      fp = fopen(filenames[numfile], "r");
      if (fp == NULL)
      {
         printf("Unable to open %s\n", filenames[numfile]);
         exit(1);
      }

      presen = 0;
      cursen = 0;

      cptr = fgets(buffer, BUFFERSIZE, fp);
      while (cptr != NULL)  // Process each sensor event in the file
      {
         sscanf(buffer, "%s %s %s %s %s %s",
                   sdate, stime, sensorid, newsensorid, sensorstatus, alabel);
         cursen = FindSensor(newsensorid);

         // Update the MI count based on the temporal proximity of presen
         // (the previous sensor) to cursen (the current sensor)
         MI[presen][cursen]++;
         presen = cursen;
         cptr = fgets(buffer, BUFFERSIZE, fp);
      }
      fclose(fp);
   }

   for (i=0; i<NumSensors; i++)  // Normalize MI values for each pair of sensors
   {
      total = 0;
      for (j=0; j<NumSensors; j++)
         total += MI[i][j];
      if (total != 0)
      {
         for (j=0; j<NumSensors; j++)
            MI[i][j] /= total;
         if (MI[i][j] == 0.0)
	    MI[i][j] = MIN;
      }
   }
}


/* Read the data from the input files.  Compute a feature vector for
   each window-length sequence of sensor events. */
void ReadData()
{
   FILE *fp;
   char *cptr=NULL;
   char buffer[BUFFERSIZE], date[MAXSTR], stime[MAXSTR], sensorid[MAXSTR];
   char newsensorid[MAXSTR], sensorstatus[MAXSTR], alabel[MAXSTR];
   char prevdate[MAXSTR];
   int i, numfile, end, snum;
   float *tempdata;

                               // Feature length is #sensors*d + 12 time values
   NumFeatures = (5*NumSensors) + 6;

   tempdata = (float *) malloc(NumFeatures * sizeof(float));
   if (tempdata == NULL)
   {
      fprintf(stderr, "Memory error in ReadData\n");
      exit(1);
   }

   for (numfile=0; numfile<NumFiles; numfile++)  // Process each data file
   {
      strcpy(date, "");
      if ((OutputLevel > 0) && (Mode != TEST))
         printf("   Reading file %s...", filenames[numfile]);
      fp = fopen(filenames[numfile], "r");
      if (fp == NULL)
      {
         printf("Unable to open %s", filenames[numfile]);
         exit(1);
      }

      cptr = fgets(buffer, BUFFERSIZE, fp);
      while (cptr != NULL)  // Process each sensor event in the file
      {
	 strcpy(prevdate, date);
         sscanf(buffer, "%s %s %s %s %s %s",
                date, stime, sensorid, newsensorid, sensorstatus, alabel);
	 if (strcmp(prevdate, date) != 0) // new day
	    for (i=0; i<NumSensors; i++)
	       sensortimes[i] = 86400;
	 snum = FindSensor(sensorid);
	 sensortimes[snum] = ComputeSeconds(stime);   // last time sensor fired
         // initialize the temp data
         for (i=0; i<NumFeatures; i++)
            tempdata[i] = 0;

	                       // Compute a feature vector for the sensor event
	 if ((IgnoreOther == FALSE) || (strcmp(alabel, "Other_Activity") != 0))
            end = ComputeFeature(stime, sensorid, newsensorid, tempdata);
	 else end = 0;
         if (end == 1)            // End of window reached, add feature vector
         {
            if (NumData == 0)
	    {
               data = (float **) malloc(sizeof(float *));
               if (data == NULL)
               {
                  fprintf(stderr, "Memory error in ReadData\n");
                  exit(1);
               }
               labels = (int *) malloc(sizeof(int));
               if (labels == NULL)
               {
                  fprintf(stderr, "Memory error in ReadData\n");
                  exit(1);
               }
	    }
            else
	    {
	       data = (float **) realloc(data, (NumData+1) * sizeof(float *));
               if (data == NULL)
               {
                  fprintf(stderr, "Memory error in ReadData\n");
                  exit(1);
               }
               labels = (int *) realloc(labels, (NumData+1) * sizeof(int));
               if (labels == NULL)
               {
                  fprintf(stderr, "Memory error in ReadData\n");
                  exit(1);
               }
	    }
            data[NumData] = (float *) malloc(NumFeatures * sizeof(float));
            if (data[NumData] == NULL)
            {
               fprintf(stderr, "Memory error in ReadData\n");
               exit(1);
            }
            labels[NumData] = FindActivity(alabel);
	    classfreq[labels[NumData]] += 1;

            for (i=0; i<NumFeatures; i++)   // Store data point
               data[NumData][i] = tempdata[i];

            NumData++;
         }
         cptr = fgets(buffer, BUFFERSIZE, fp);
      }
      fclose(fp);

      if ((OutputLevel > 0) && (Mode != TEST))
         printf("done\n");
   }

   free(tempdata);
}


/* Compute the feature vector for each window-size sequence of sensor events.
   fields
   0 -  3: quantized value for window times
   4:      window size in number of sensor events
   5:      window size in time duration
   6 - NumSensors+6: count of events for each unique sensor1
   NumSensors+6 - 2*NumSensors+6: count of events for each unique sensor2
   2*NumSensors+6 - 3*NumSensors+6: dominant sensor for previous window
   3*NumSensors+6 - 4*NumSensors+6: dominant sensor two windows back
   4*NumSensors+6 - 5*NumSensors+6: time since sensor last fired (<= 86400)*/
int ComputeFeature(char *time, char *sensorid1, char *sensorid2, float *tempdata)
{
   int i, wsize, duration, max=0, *scount, ctime, diff;

   windata[wincnt][HOUR] = ComputeHour(time);
   windata[wincnt][SENSOR1] = FindSensor(sensorid1);
   windata[wincnt][SENSOR2] = FindSensor(sensorid2);
   ctime = ComputeSeconds(time);
   windata[wincnt][SECONDS] = ctime;

   if (wincnt < (MAXWINDOW-1))  // Not reached end of window
   {
      wincnt++;  // Keep track of number of events in current window
      return(0);
   }
   else                      // Reached end of window, create feature vector
   {
      wsize = GetWindow(sensorid1);             // determine the window size

      scount = (int *) malloc(NumSensors * sizeof(int));
      if (scount == NULL)
      {
         fprintf(stderr, "Memory error in ComputeFeature\n");
         exit(1);
      }
      for (i=0; i<NumSensors; i++)
         scount[i] = 0;

                               // Determine the dominant sensor for this window
      for (i=(MAXWINDOW-1); i>=(MAXWINDOW-wsize); i--)
      {
         scount[windata[i][SENSOR1]]++;
         scount[windata[i][SENSOR2]]++;
      }
      if ((numwin % MAXWINDOW) == 0)
      {
         prevwin2 = prevwin1;
         prevwin1 = dominant;
         dominant = 0;
         for (i=0; i<NumSensors; i++)
            if (scount[i] > max)
	    {
	       max = scount[i];
	       dominant = i;
	    }
      }
      free(scount);

      // Sensor attributes
      // Weight each sensor event within the window
      // according to the selected weight scheme (none, time, MI, or time+MI)
      for (i=(MAXWINDOW-1); i>=(MAXWINDOW-wsize); i--)
      {
         if (Weight == NONE)                       // Increment sensor count
	 {
            tempdata[windata[i][SENSOR1]+6]++;
            tempdata[windata[i][SENSOR2]+6+NumSensors]++;
	 }

         else if (Weight == TIMEWEIGHT)
	 {
            tempdata[windata[i][SENSOR1]+6] += (float) exp(-KAPPA *
                    fabs(windata[MAXWINDOW-1][SECONDS] - windata[i][SECONDS]));
            tempdata[windata[i][SENSOR2]+6+NumSensors] += (float) exp(-KAPPA *
                    fabs(windata[MAXWINDOW-1][SECONDS] - windata[i][SECONDS]));
         }

         else if (Weight == MUTUALINFO)
	 {
            tempdata[windata[i][SENSOR1]+6] +=
                 MI[windata[MAXWINDOW-1][SENSOR1]][windata[wincnt][SENSOR1]];
            tempdata[windata[i][SENSOR2]+6+NumSensors] +=
                 MI[windata[MAXWINDOW-1][SENSOR2]][windata[wincnt][SENSOR2]];
         }

         else if (Weight == MUTUALINFO_TIME)
	 {
            tempdata[windata[i][SENSOR1]+6] += (float) exp(-KAPPA *
                  fabs(windata[MAXWINDOW-1][SECONDS] - windata[i][SECONDS])) *
                  MI[windata[MAXWINDOW-1][SENSOR1]][windata[wincnt][SENSOR1]];
            tempdata[windata[i][SENSOR2]+6+NumSensors] += (float) exp(-KAPPA *
                  fabs(windata[MAXWINDOW-1][SECONDS] - windata[i][SECONDS])) *
                  MI[windata[MAXWINDOW-1][SENSOR2]][windata[wincnt][SENSOR2]];
         }

         // Normalize the sensor counts to not be affected by window size
	 if ((Weight != TIMEWEIGHT) && (Weight != MUTUALINFO_TIME))
	 {
	    tempdata[windata[i][SENSOR1]+6] /= (float) wsize;
	    tempdata[windata[i][SENSOR2]+6] /= (float) wsize;
         }
      }

      // time attributes
      for (i=(MAXWINDOW-1); i>=(MAXWINDOW-wsize); i--)
         tempdata[windata[i][HOUR]/6] += (float) 1 / (float) wsize;

      tempdata[4] = wsize;    // size of window
      duration =
         abs(windata[MAXWINDOW-wsize][HOUR] - windata[MAXWINDOW-1][HOUR]);
      tempdata[5] = duration; // time duration of window

      // previous windows
      for (i=0; i<NumSensors; i++)
      {
         if ((i == prevwin1) && (numwin > MAXWINDOW))
	    tempdata[6+(2*NumSensors)+i] = 1.0;
	 else tempdata[6+(2*NumSensors)+i] = 0.0;
         if ((i == prevwin2) && (numwin > (MAXWINDOW*2)))
	    tempdata[6+(3*NumSensors)+i] = 1.0;
	 else tempdata[6+(3*NumSensors)+i] = 0.0;
      }

      // time offset since each sensor fired
      for (i=0; i<NumSensors; i++)
      {
         diff = ctime - sensortimes[i];
	 if (diff < 86400)
	    diff += 86400;
         if (diff > 86400)
	    diff = 86400;
	 tempdata[6+(4*NumSensors)+i] = (float) diff / (float) 86400;
      }

      for (i=0; i<MAXWINDOW-1; i++)
      {
         windata[i][HOUR] = windata[i+1][HOUR];
         windata[i][SENSOR1] = windata[i+1][SENSOR1];
         windata[i][SENSOR2] = windata[i+1][SENSOR2];
         windata[i][SECONDS] = windata[i+1][SECONDS];
      }
      numwin++;
      return(1);
   }

   return(0);
}


/* Compute the hour of the day (based on 24 hour clock). */
int ComputeHour(char *time)
{
   char temp[3];

   temp[0] = time[0];
   temp[1] = time[1];
   temp[2] = '\0';

   return(atoi(temp));
}


/* Compute the number of seconds past midnight on the current day. */
int ComputeSeconds(char *time)
{
   char hour[3], minute[3], second[3];

   hour[0] = time[0];
   hour[1] = time[1];
   hour[2] = '\0';
   minute[0] = time[3];
   minute[1] = time[4];
   minute[2] = '\0';
   second[0] = time[6];
   second[1] = time[7];
   second[2] = '\0';

   return((atoi(hour)*3600)+(atoi(minute)*60)+atoi(second));
}


/* Compute the window size for a given sensor. */
int GetWindow(char *name)
{
   int win;

   win = factwind[fsenact[FindSensor(name)]];

   if (win == 0)
      return 5;
   else if (win == 1)
      return 10;
   else if (win == 2)
      return 20;
   else return 30;
}


/* Create a decision tree model for every node in the hierarchy. */
void TrainModel()
{
   FILE *fpdata;
   char str[MAXSTR];
   int i, j;

   if (OutputLevel > 0)
      printf("   Training the model\n");
   sprintf(str, "%s/model", modelfname);

   if (WRITEDATA == TRUE) // Write data points to a file in arff format
   {
      fpdata = fopen("ar.arff", "w");
      fprintf(fpdata, "@relation ar\n\n");
      for (i=0; i<NumFeatures; i++)
         fprintf(fpdata, "@attribute A%d numeric\n", i);
      fprintf(fpdata, "@attribute class {");
      for (i=0; i<NumActivities; i++)
         if (i < (NumActivities-1))
            fprintf(fpdata, "%d,", i);
         else fprintf(fpdata, "%d}\n\n@data\n", i);

      for (i=0; i<NumData; i++)
      {
         for (j=0; j<NumFeatures; j++)
            fprintf(fpdata, "%f,", data[i][j]);
         fprintf(fpdata, "%d\n", labels[i]);
      }
      fclose(fpdata);
   }
   else DTL();       // train the decision tree
   if (OutputLevel > 0)
      printf("   done\n");
}


void DTL()
{
   int i, j, majorclass, *training, numtrain, small;
   TreeNodePtr dt;

   numtrain = NumData;
   training = (int *) calloc(numtrain, sizeof(int));
   for (i=0; i<numtrain; i++)
      training[i] = i;

   attused = (int *) malloc(NumFeatures * sizeof(int));
   for (i=0; i<NumFeatures; i++)
      attused[i] = FALSE;

   small = FALSE;
   majorclass = MajorityClass(training, NumData, &small);
   if (NumData > 0)
   {
      dt = BuildTree(training, numtrain, majorclass, attused,
                     NumFeatures, small);
      if (PRUNE == TRUE)
      {
         if (OutputLevel > 0)
            printf("      Pruning...");
         dt = PruneTree(dt);
         if (OutputLevel > 0)
            printf("done\n");
      }
   }

   // save the model for the current node using the provided name
   SaveDT(dt);

   free(training);
   free(attused);
   FreeTree(dt);
}


/* Prune decision tree based on estimated error.
   if E(leaf) <= (E(tree)+0.1) and E(leaf) <= (E(largest branch)+0.1)
   then collapse subtree into leaf. */
TreeNodePtr PruneTree(TreeNodePtr tree)
{
   int i, num;
   float etree, eleaf, ebranch;

   if (tree->numChildren == 0)
      return(tree);

   else
   {
      tree->children[0] = PruneTree(tree->children[0]);
      tree->children[1] = PruneTree(tree->children[1]);
      num = NumInClass(tree, tree->Class, TREE);
      etree = 1.0 - ((float) num / (float) tree->numInstances);
      etree = AddErr((float) NumData, etree);

      num = NumInClass(tree, tree->Class, LEAF);
      eleaf = 1.0 - ((float) num / (float) tree->numInstances);
      eleaf = AddErr((float) NumData, eleaf);

      if (tree->children[0]->numInstances > tree->children[1]->numInstances)
      {
         num = NumInClass(tree->children[0], tree->children[0]->Class, TREE);
         ebranch = 1.0 -
	    ((float) num / (float) tree->children[0]->numInstances);
         ebranch = AddErr((float) tree->children[0]->numInstances, ebranch);
      }
      else
      {
         num = NumInClass(tree->children[1], tree->children[1]->Class, TREE);
         ebranch = 1.0 -
	    ((float) num / (float) tree->children[1]->numInstances);
         ebranch = AddErr((float) tree->children[1]->numInstances, ebranch);
      }

      // if meet criteria, collapse subtree into leaf
      //if ((eleaf <= (etree + 0.1)) && (eleaf <= (ebranch + 0.1)))
      if ((eleaf <= etree) && (eleaf <= ebranch))
      {
         FreeTree(tree->children[0]);
         FreeTree(tree->children[1]);
         tree->numChildren = 0;
         tree->children = NULL;
      }
   }

   return(tree);
}


/* Determine the number of data points represented by a particular subtree
   that would be categorized correctly using majorclass or the class value
   indicated at the leaf node. */
int NumInClass(TreeNodePtr tree, int majorclass, int type)
{
   int i, num = 0;

   if (tree->numChildren == 0)   // Leaf node
   {
      for (i=0; i<tree->numInstances; i++)
         if (type == LEAF)      // Accuracy if subtree collapsed into leaf node
	 {
            if (labels[tree->instances[i]] == majorclass)
	       num++;
         }
	 else                   // Accuracy if subtree kept
	 {
            if (labels[tree->instances[i]] == tree->Class)
	       num++;
	 }
   }
   else num = NumInClass(tree->children[0], majorclass, type) +
              NumInClass(tree->children[1], majorclass, type);

   return(num);
}


/*  Compute the additional errors if the error rate increases to the
    upper limit of the confidence level.  The coefficient is the
    square of the number of standard deviations corresponding to the
    selected confidence level.  (Taken from Documenta Geigy Scientific
    Tables (Sixth Edition), page 185 (with modifications).) */
float AddErr(float n, float e)
{
   int i;
   float c=0.0, val0, pr;

   i = 0;
   // Compute coefficient value, interprolating from values in Val and Dev
   while (CF > Val[i])
      i++;
   c = Dev[i-1] +
       (Dev[i] - Dev[i-1]) * (CF - Val[i-1]) /(Val[i] - Val[i-1]);
   c = c * c;

   if (e < 0.000001)
      return(n * (1.0 - expf(logf(CF) / n)));
   else if (e < 0.9999)
   {
      val0 = n * (1.0 - expf(logf(CF) / n));
      return(val0 + (e * (AddErr(n, 1.0) - val0)));
   }
   else if ((e + 0.5) >= n)
      return(0.67 * (n - e));
   else
   {
      pr = (e + 0.5 + (c / 2.0)
              + sqrt(c * ((e + 0.5) * (1 - (e + 0.5)/n) + (c / 4.0)))) /
	   (n + c);
      return((n * pr) - e);
   }
}


/* Recursively build decision tree. */
TreeNodePtr BuildTree(int *training, int numtrain, int parentclass,
                      int *attused, int numatt, int small)
{
   TreeNodePtr tree;
   int i, att, majorclass, newtrain, tmp, num;
   float val;

   tree = AllocateTreeNode();
   tree->numInstances = numtrain;    // Look at subset of data for this subtree
   tree->instances = (int *) malloc(numtrain * sizeof(int));
   for (i=0; i<numtrain; i++)
      tree->instances[i] = training[i];
   if (numtrain == 0)                // No data left, make this a leaf node
      tree->Class = parentclass;
   else
   {
      majorclass = MajorityClass(training, numtrain, &small);
      tree->Class = majorclass;

      // If all of the data belongs to the same class or the number of
      // data points for a child is too small, make this a leaf node
      if  ((AllOneClass(training, numtrain) == FALSE) &&
	   (small == FALSE))  // C4.5
      {
	 if (C45 == TRUE) // Identify the best attribute and threshold to split
            num = SelectC45Attribute(tree, training, numtrain, attused);
         else num = SelectBestAttribute(tree, training, numtrain, attused);
	 if (num > 1)        // Split the data and make subtrees for each child
	 {
	    att = tree->attribute;
	    attused[att] = TRUE;     // Do not use attribute again in same path
	    tree->numChildren = 2;
	    tree->children = (TreeNodePtr *) calloc(2, sizeof(TreeNodePtr));
	    val = tree->threshold;
	    newtrain = CompactInst(training, numtrain, att, val);

	    // Build left subtree
	    tree->children[0] = BuildTree(training, newtrain, majorclass,
	                                  attused, numatt, small);

	    // swap instances on either side of threshold
            for (i=0; i<newtrain; i++)
	    {
	       numtrain--;
	       tmp = training[i];
	       training[i] = training[numtrain];
	       training[numtrain] = tmp;
	    }

	    // Build right subtree
	    tree->children[1] = BuildTree(training, numtrain, majorclass,
	                                  attused, numatt, small);
	    attused[att] = FALSE;  // Can reuse attribute in other part of tree

	                                            // Online pruning criteria
	    if ((tree->children[0]->numInstances <= MINNUMOBJ) ||
	        (tree->children[1]->numInstances <= MINNUMOBJ))
	    {
	       FreeTree(tree->children[0]);
	       FreeTree(tree->children[1]);
	       tree->numChildren = 0;
	       tree->children = NULL;
	       tree->Class = majorclass;
	    }
         }
      }
   }

   return(tree);
}


/* Sets the attribute field of dtnode and the threshold
   according to the unused attribute having the best split potential.
   The split potential is measured as greatest information gain. */
int SelectC45Attribute(TreeNodePtr dtnode, int *training, int numtrain,
                       int *attused)
{
   int i, att, bestatt, numthresholds, numatt=0;
   float *thresholds, defaultent;
   float gain=0.0, bestthreshold, bestgain=0.0, avggain=0.0;

   defaultent = OldEnt(training, numtrain);       // Entropy of the parent node
   for (att=0; att<NumFeatures; att++)
   {
      if (attused[att] == FALSE)         // Do not reuse attribute on same path
      {
	 // Calculate set of possible thresholds for this attribute
         thresholds =
	    ComputeC45Thresholds(training, numtrain, att, &numthresholds);

         for (i=0; i<numthresholds; i++)  // Test each attribute
	 {
	    gain = defaultent - NewEnt(training, numtrain, att, thresholds[i]);
	    if (gain > bestgain)     // Find attribute that yields maximum gain
	    {
	       bestatt = att;
	       bestthreshold = thresholds[i];
	       bestgain = gain;
	    }
	    if (gain > 0.0)          // Do not split if no gain
	    {
	       numatt++;
	       avggain += gain;
	    }
	 }
         if (OutputLevel > 1)
            printf("   att %d numt %d default %f gain %f\n",
	           att, numthresholds, defaultent, gain);
      }
   }

   if (numatt == 0)
      avggain = 0.0;
   else avggain /= (float) numatt;
   //if (bestgain > (avggain - 0.001))
   if (bestgain > avggain)
   {
      if (OutputLevel > 1)
         printf("best %d threshold %f\n", bestatt, bestthreshold);
      dtnode->attribute = bestatt;
      dtnode->threshold = bestthreshold;
      return(2);
   }
   else return(0);  // No gain, do not split
}



/* Sets the attribute field of dtnode (and possibly the threshold too)
   according to the unused attribute having the best split potential.
   The split potential is measured as (Langley)
   the number of training examples at the current node that would be correctly
   classified by the children nodes created by splitting on that attribute. */
int SelectBestAttribute(TreeNodePtr dtnode, int *training, int numtrain,
                        int *attused)
{
   int i, att, bestatt = 0, numthresholds, accuracy, bestval = 0, total;
   float *thresholds, bestthreshold = 0.0, val, gain;

   total = 0;
   for (att=0; att<NumFeatures; att++)
      if (attused[att] == FALSE)
      {
         thresholds =
	    ComputeThresholds(training, numtrain, att, &numthresholds);
         for (i=0; i<numthresholds; i++)
         {
	    val = thresholds[i];
	    accuracy = AttMajorClass(training, numtrain, att, val, LT);
	    accuracy += AttMajorClass(training, numtrain, att, val, GE);
	    if (accuracy >= bestval)
	    {
	       bestatt = att;
	       bestthreshold = thresholds[i];
	       bestval = accuracy;
	    }
         }
      total += numthresholds;
      free(thresholds);
   }
   dtnode->attribute = bestatt;
   dtnode->threshold = bestthreshold;

   return(total);
}


/* Calculate the entropy of a set of data based on the distribution of class
   labels found in the data. */
float OldEnt(int *training, int numtrain)
{
   int i, *freq;
   float e;

   freq = (int *) calloc(NumActivities, sizeof(int));
   if (freq == NULL)
   {
      printf("Memory error in OldEnt\n");
      exit(1);
   }
   for (i=0; i<numtrain; i++)
      freq[labels[training[i]]] += 1;

   e = Entropy(freq, numtrain);
   free(freq);

   return(e);
}


/* Calculate the weighted entropy of a node based on the frequencies
   of the class labels found in the children of this node. */
float NewEnt(int *training, int numtrain, int att, float val)
{
   int i, *freq1, *freq2, total1=0, total2=0;
   float e, e1, e2;

   freq1 = (int *) calloc(NumActivities, sizeof(int));
   freq2 = (int *) calloc(NumActivities, sizeof(int));
   if ((freq1 == NULL) || (freq2 == NULL))
   {
      printf("Memory error in NewEnt\n");
      exit(1);
   }
   for (i=0; i<numtrain; i++)
      if (data[training[i]][att] < val)
      {
         freq1[labels[training[i]]] += 1;
	 total1 += 1;
      }
      else
      {
         freq2[labels[training[i]]] += 1;
	 total2 += 1;
      }

   e1 = Entropy(freq1, total1);
   e2 = Entropy(freq2, total2);
   free(freq1);
   free(freq2);
   e = ((total1 / (float) numtrain) * e1) + ((total2 / (float) numtrain) * e2);

   return(e);
}


/* Calculate entropy of a set of values. */
float Entropy(int *freq, int total)
{
   int i;
   float sum = 0.0, p;

   for (i=0; i<NumActivities; i++)
   {
      if (total == 0)
         p = 0.0;
      else p = (float) freq[i] / (float) total;
      if (p > 0.000001)
         sum -= p * log2f(p);
   }

   return(sum);
}


/* Generate the set of class labels that correspond to a set of data points. */
int FindClass(int *training, int numtrain, int att, float value)
{
   int i;

   for (i=0; i<numtrain; i++)
      if (data[training[i]][att] == value)
         return(labels[training[i]]);

   return(0);
}


/* Compare x with y to use in sort function. */
int compfn(const void *v1, const void *v2)
{
   float *x = (float *) v1;
   float *y = (float *) v2;

   if (*x < *y)
      return(-1);
   if (*x > *y)
      return(1);
   return(0);
}


/* Returns an array of thresholds determined from attribute attr in
   the first numTrain instances in training.  After determining the
   sorted sequence of unique values <v1,v2,...,vn> for attr, the n-1
   thresholds are <(v1+v2)/2,...,(vn-1+vn)/2>.  numThresholds is
   called by reference to return the number of such thresholds. */
float *ComputeC45Thresholds(int *training, int numtrain, int att,
                         int *numthresholds)
{
   int i, n, v, minsplit, c1, c2;
   float v1, v2, *values;

   minsplit = (int) (0.1 * (float) numtrain / (float) NumActivities);
   minsplit = max(min(minsplit, 25), MINNUMOBJ);
   values = (float *) calloc(numtrain, sizeof(float));
   for (i=0; i<numtrain; i++)
      values[i] = data[training[i]][att];
   qsort(values, numtrain, sizeof(float), compfn);
   v1 = values[0];
   n = 0;
   v = 1;
   while (v < numtrain)
   {
      v2 = values[v];
      // Only consider a cut point if big enough gap and data points
      // belong to different classes (Fayyad and Irani, 1993)
      c1 = FindClass(training, numtrain, att, v1);
      c2 = FindClass(training, numtrain, att, v2);
      if (((v1 + 0.00001) < v2) &&
          (v >= minsplit) && ((numtrain - v) >= minsplit) &&
	  (c1 != c2))
      {
         values[n++] = (v1 + v2) / 2.0;  // threshold is midway between values
	 v1 = v2;
      }
      v++;
   }
   *numthresholds = n;

   return(values);
}


/* Returns an array of thresholds determined from attribute attr in
   the first numTrain instances in training.  After determining the
   sorted sequence of unique values <v1,v2,...,vn> for attr, the n-1
   thresholds are <(v1+v2)/2,...,(vn-1+vn)/2>.  numThresholds is
   called by reference to return the number of such thresholds. */
float *ComputeThresholds(int *training, int numtrain, int att,
                         int *numthresholds)
{
   int i, n, v;
   float v1, v2, *values;

   values = (float *) calloc(numtrain, sizeof(float));
   for (i=0; i<numtrain; i++)
      values[i] = data[training[i]][att];
   qsort(values, numtrain, sizeof(float), compfn);
   v1 = values[0];
   n = 0;
   v = 1;
   while (v < numtrain)
   {
      v2 = values[v];
      if ((v1 + 0.00001) < v2) // C4.5
      {
         values[n++] = (v1 + v2) / 2.0;
	 v1 = v2;
      }
      v++;
   }
   *numthresholds = n;

   return(values);
}


/* Moves all instances in training having attribute attr equal to (or
   less than if numeric) value val to the left of the training array
   and returns the number of such instances. */
int CompactInst(int *training, int numtrain, int att, float val)
{
   int left, right, tmp;

   left = 0;
   right = numtrain - 1;
   while (left < right)
   {
      while ((left < numtrain) && (data[training[left]][att] < val))
         left++;
      while ((right > 0) && (data[training[right]][att] >= val))
         right--;
      if (left < right)
      {
         tmp = training[left];
	 training[left] = training[right];
	 training[right] = tmp;
      }
   }

   return(left);
}


/* Return TRUE if all instances have the same class. */
int AllOneClass(int *training, int numtrain)
{
   int i, firstclass;

   firstclass = labels[training[0]];
   for (i=1; i<numtrain; i++)
      if (labels[training[i]] != firstclass)
         return(FALSE);
   return(TRUE);
}


/* Allocate memory for a node in the decision tree. */
TreeNodePtr AllocateTreeNode()
{
   TreeNodePtr node;

   node = (TreeNodePtr) malloc(sizeof(TreeNode));
   node->attribute = 0;
   node->threshold = 0.0;
   node->numChildren = 0;
   node->children = NULL;
   node->Class = 0;
   node->numInstances = 0;

   return(node);
}


/* Return the class label that is most frequent for a set of data points. */
int MajorityClass(int *training, int num, int *small)
{
   int i, *cfreq, max, maxfreq;

   cfreq = (int *) malloc(NumActivities * sizeof(int));

   for (i=0; i<NumActivities; i++)
      cfreq[i] = 0;

   for (i=0; i<num; i++)
      cfreq[labels[training[i]]] += 1;

   max = 0;
   maxfreq = cfreq[0];

   for (i=1; i<NumActivities; i++)
   {
      if (cfreq[i] > maxfreq)
      {
         max = i;
	 maxfreq = cfreq[i];
      }
   }
   if (num < (float) (2 * MINNUMOBJ)) // Need at least MINNUMOBJ for a leaf
      *small = TRUE;

   free(cfreq);

   return(max);
}


/* Returns the number of instances in the majority class of those instances
   with att's value comparable to val according to the test. */
int AttMajorClass(int *training, int numtrain, int att, float val, int test)
{
   int i, nummajclass;

   for (i=0; i<NumActivities; i++)
      classfreq[i] = 0;
   for (i=0; i<numtrain; i++)
      switch(test)
      {
         case LT:
	    if (data[training[i]][att] < val)
	       classfreq[labels[training[i]]] += 1;
            break;
         case GE:
	    if (data[training[i]][att] >= val)
	       classfreq[labels[training[i]]] += 1;
            break;
      }

   nummajclass = classfreq[0];
   for (i=1; i<NumActivities; i++)
      if (classfreq[i] > nummajclass)
         nummajclass = classfreq[i];

   return(nummajclass);
}


/* Free memory for a decision tree. */
void FreeTree(TreeNodePtr tree)
{
   int i;

   if (tree->numChildren > 0)
   {
      for (i=0; i<tree->numChildren; i++)
         FreeTree(tree->children[i]);
      free(tree->children);
   }
   if (tree->numInstances > 0)
      free(tree->instances);
   free(tree);
}


/* Test the decision tree model for every node in the hierarchy. */
void TestModel()
{
   int i, j, att, cval;
   TreeNodePtr dt;

   dt = LoadModel();    // load the regular model

   for (i=0; i<NumData; i++)
   {
      if (OutputLevel > 1)
         printf("Classify %d ", i);
      cval = DTLClassify(i, dt); // Use decision tree to output value

      if (cval == labels[i])
         numright++;
      if (OutputLevel > 1)
         printf("%d %d\n", labels[i], cval);
   }
}


/* Return class found at leaf of tree by following path according to the
   attribute values of the instance. */
int DTLClassify(int instance, TreeNodePtr tree)
{
   int att;

   while (tree->numChildren > 0)
   {
      att = tree->attribute;
      if (data[instance][att] < tree->threshold)
         tree = tree->children[0];
      else tree = tree->children[1];
   }

   return(tree->Class);
}


/* Save the decision tree to a file. */
void SaveDT(TreeNodePtr dt)
{
   FILE *fpmodel;
   char str[MAXSTR];

   printf("   Saving the model...");
   sprintf(str, "%s/model", modelfname);
   fpmodel = fopen(str, "w");
   if (fpmodel == NULL)
   {
      printf("File %s not writeable\n", str);
      exit(1);
   }
   WriteDT(fpmodel, dt);
   fclose(fpmodel);
   printf("done\n");
}


/* Write the parameters describing a subtree in the decision tree to a file. */
void WriteDT(FILE *fpmodel, TreeNodePtr dt)
{
   int att;

   if (dt != NULL)
   {
      fprintf(fpmodel, "%d %d\n", dt->numChildren, dt->numInstances);
      if (dt->numChildren == 0)
         fprintf(fpmodel, "%d\n", dt->Class);
      else
      {
         att = dt->attribute;
	 fprintf(fpmodel, "%d %f\n", att, dt->threshold);
	 WriteDT(fpmodel, dt->children[0]);
	 WriteDT(fpmodel, dt->children[1]);
      }
   }
}


/* Read decision tree information from a file and store it as a
   decision tree. */
TreeNodePtr ReadDT(FILE *fpmodel)
{
   TreeNodePtr dt;

   dt = AllocateTreeNode();
   fscanf(fpmodel, "%d %d\n", &(dt->numChildren), &(dt->numInstances));
   if (dt->numChildren == 0)
      fscanf(fpmodel, "%d\n", &(dt->Class));
   else
   {
      fscanf(fpmodel, "%d %f\n", &(dt->attribute), &(dt->threshold));
      dt->children = (TreeNodePtr *) calloc(2, sizeof(TreeNodePtr));
      dt->children[0] = ReadDT(fpmodel);
      dt->children[1] = ReadDT(fpmodel);
   }

   return(dt);
}


/* Read the decision tree. */
TreeNodePtr LoadModel()
{
   FILE *fpmodel;
   char str[MAXSTR];
   TreeNodePtr dt;

   sprintf(str, "%s/model", modelfname);
   printf("Reading model from file %s...\n", str);
   fpmodel = fopen(str, "r");
   if (fpmodel == NULL)
   {
      printf("File %s not writeable\n", str);
      exit(1);
   }
   dt = ReadDT(fpmodel);
   fclose(fpmodel);
   printf("done\n");

   return(dt);
}


/* Free allocated memory. */
void FreeMemory()
{
   int i;

   if (NumFiles > 0)
   {
      for (i=0; i<NumFiles; i++)
         free(filenames[i]);
      free(filenames);
   }
   if (modelfname != NULL)
      free(modelfname);
   if (NumSensors > 0)
   {
      for (i=0; i<NumSensors; i++)
         free(sensornames[i]);
      free(sensornames);
      free(sensortimes);
   }
   if (senact != NULL)
   {
      for (i=0; i<NumSensors; i++)
         free(senact[i]);
      free(senact);
   }
   if (actwind != NULL)
      free(actwind);
   if (factwind != NULL)
      free(factwind);
   if (fsenact != NULL)
      free(fsenact);
   if (MI != NULL)
   {
      for (i=0; i<NumSensors; i++)
         free(MI[i]);
      free(MI);
   }
   if (NumData > 0)
   {
      for (i=0; i<NumData; i++)
         free(data[i]);
      free(data);
      free(labels);
   }
   if (NumActivities > 0)
   {
      for (i=0; i<NumActivities; i++)
         free(activitynames[i]);
      free(activitynames);
      free(classfreq);
   }
}
